/* de2022cn.c  -- ISO-2022-CN (RFC1922) decoder
   Copyright (C) 1995, 1996 Electrotechnical Laboratory, JAPAN.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


/* Author: K.Handa <handa@etl.go.jp>
   History:
	1996.4.6 -- Version 0.9beta released.
*/

#include <stdio.h>
#include "iso2022.h"

/* State of locking-shift.  0 means in SI, 1 means in SO.  */
int shiftout;

/* Table of <final-char> of character sets designated to each graphic
   registers.  The value -1 means nothing designated.  */
char REG[4];

/* Macros for indexing the array REG.  */
#define G0	0
#define G1	1
#define G2	2
#define G3	3

enum target_code {
  gb,
  cns,
  big5
};

void
initialize ()
{
  int i;

  /* Initialize context of ISO-2022.  */
  shiftout = 0;
  REG[G0] = CHARSET_ASCII;
  REG[G1] = -1;
  REG[G2] = -1;
  REG[G3] = -1;
}

int produce_target_code (target, buf, charset, c1, c2)
     enum target_code target;
     unsigned char *buf;
     char charset;
     unsigned char c1, c2;
{
  unsigned char *buf_start = buf;
  unsigned int b5;

  if (target == gb)
    {
      switch (charset)
	{
	case CHARSET_GB2312:
	  break;

	case CHARSET_GB7589:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA2;
	  break;

	case CHARSET_GB7590:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA3;
	  break;

	case CHARSET_CNS11643_1:
	case CHARSET_CNS11643_2:
	case CHARSET_CNS11643_3:
	case CHARSET_CNS11643_4:
	  if (b5 = ISO_TO_BIG5 (charset, c1, c2))
	    {
	      unsigned int high = b5 >> 8, low = b5 & 0xFF;
	      unsigned int gbcode = BIG5_TO_GB (high, low);

	      if (!gbcode) gbcode = BLANK_SQUARE_GB2312;
	      c1 = gbcode >> 8, c2 = gbcode & 0xFF;
	      break;
	    }
	  /* fall through ... */
	default:
	  c1 = BLANK_SQUARE_GB2312 >> 8;
	  c2 = BLANK_SQUARE_GB2312 & 0xFF;
	}
      *buf++ = c1 | 0x80;
      *buf++ = c2 | 0x80;
    }
  else if (target == cns)
    {
    try_cns_again:
      switch (charset)
	{
	case CHARSET_CNS11643_1:
	  break;

	case CHARSET_CNS11643_2:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA2;
	  break;

	case CHARSET_CNS11643_3:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA3;
	  break;

	case CHARSET_CNS11643_4:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA4;
	  break;

	case CHARSET_CNS11643_5:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA5;
	  break;

	case CHARSET_CNS11643_6:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA6;
	  break;

	case CHARSET_CNS11643_7:
	  *buf++ = ISO_SS2_8;
	  *buf++ = 0xA7;
	  break;

	case CHARSET_GB2312:
	  if (b5 = ISO_TO_BIG5 (charset, c1, c2))
	    {
	      unsigned int high = b5 >> 8, low = b5 & 0xFF;
	      unsigned int cnscode;

	      BIG5_TO_CNS (high, low, charset, cnscode);
	      if (charset)
		{
		  c1 = cnscode >> 8, c2 = cnscode & 0xFF;
		  goto try_cns_again;
		}
	      charset = CHARSET_GB2312;
	    }
	  /* fall through ... */
	default:
	  c1 = BLANK_SQUARE_CNS11643_1 >> 8;
	  c2 = BLANK_SQUARE_CNS11643_1 & 0xFF;
	}
      *buf++ = c1 | 0x80;
      *buf++ = c2 | 0x80;
    }
  else				/* target == big5 */
    {
      unsigned int b5 = ISO_TO_BIG5 (charset, c1, c2);

      if (!b5) b5 = BLANK_SQUARE_BIG5;
      *buf++ = b5 >> 8;
      *buf++ = b5 & 0xFF;
    }
  
  return (buf - buf_start);

 unknown_charset:
  *buf++ = '<';
  *buf++ = charset;
  *buf++ = c1 & 0x7F;
  *buf++ = c2 & 0x7F;
  *buf++ = '>';
  
  return (buf - buf_start);
}
  
#define SAFETY_GET_SRC(c)	\
  do {			  	\
    if (src < src_end)	  	\
      c = *src++;		\
    else			\
      {			  	\
        src = src_base;	  	\
        goto no_more_src;	\
      }			  	\
  } while (0)

int
decode (target, src, src_len, dst, dst_len, consumed)
     enum target_code target;
     unsigned char *src, *dst;
     int src_len, dst_len;
     int *consumed;
{
  register unsigned int i, c1, c2;
  unsigned char *src_start = src, *src_end = src + src_len;
  unsigned char *dst_start = dst, *dst_end = dst + dst_len;

  /* We'll produce maximally 5 bytes in one loop.  So, subtract 4 from
     `dst_end' so that overflow check is necessary only at the head of
     loop.  */
  dst_end -= 4;


  while (src < src_end && dst < dst_end)
    {
      unsigned char *src_base = src;

      c1 = *src++;
      switch (c1)
	{
	case ISO_SO:
	  shiftout = 1;
	  break;
      
	case ISO_SI:
	  shiftout = 0;
	  break;

	case ISO_ESC:
	  SAFETY_GET_SRC (c1);
	  switch (c1)
	    {
	    case '$':		/* Start of designation sequence */
	      SAFETY_GET_SRC (c1);
	      if (c1 == 'A')
		{
		  /* This is a special case for GB2312 which is
                     allowed the short form designation sequence "ESC
                     $ A" for historical reason.  */
		  REG[G0] = CHARSET_GB2312;
		}
	      else
		{
		  if (c1 < '(' || c1 > '+') goto invalid_code;
		  SAFETY_GET_SRC (c2);
		  if (c2 < '0') goto invalid_code;
		  REG[c1 - '('] = c2;
		}
	      break;

	    case ISO_SS2_7:	/* SS2 */
	      if (REG[G2] < 0) goto invalid_code;
	      SAFETY_GET_SRC (c1);
	      SAFETY_GET_SRC (c2);
	      dst += produce_target_code (target, dst, REG[G2], c1, c2);
	      break;
	      
	    case ISO_SS3_7:		/* SS3 */
	      if (REG[G3] < 0) goto invalid_code;
	      SAFETY_GET_SRC (c1);
	      SAFETY_GET_SRC (c2);
	      dst += produce_target_code (target, dst, REG[G3], c1, c2);
	      break;

	    default:
	      goto invalid_code;
	    }
	  break;

	default:
	  if (c1 <= 32 || c1 == 127)
	    /* ASCII control code.  */
	    *dst++ = c1;
	  else if (c1 < 127)
	    {
	      /* Character code.  */
	      int reg = shiftout ? 1 : 0;

	      if (REG[reg] < 0)
		/* Nothing designated to graphic regsiter REG.  */
		goto invalid_code;
	      if (REG[reg] == CHARSET_ASCII)
		*dst++ = c1;
	      else
		{
		  SAFETY_GET_SRC (c2);
		  if (c2 <=32 || c2 >= 127) goto invalid_code;
		  dst += produce_target_code (target, dst, REG[reg], c1, c2);
		}
	    }
	  else
	    /* ISO-2022-CN doesn't allow 8-bit code.  */
	    goto invalid_code;	  
	  break;

	}
      continue;

    invalid_code:
      *dst++ = *src_base++;
      src = src_base;
    }

 no_more_src:
  *consumed = src - src_start;
  return (dst - dst_start);
}

void
help (cmd)
{
  fprintf (stderr, "Usage: %s TARGET_CODE < INFILE > OUTFILE\n", cmd);
  fprintf (stderr, "\tTARGET_CODE is one of `gb', `cns', or `big5'.\n");
  fprintf (stderr, "\tINFILE is a file of ISO-2022-CN.\n");
  fprintf (stderr, "\tOUTFILE is a file decoded in TARGET_CODE.\n");
  exit (1);
}

#define INBUF_SIZE 1024
#define OUTBUF_SIZE  1024

main (argc, argv)
     int argc;
     char *argv[];
{
  unsigned char inbuf[INBUF_SIZE];
  unsigned char outbuf[OUTBUF_SIZE];
  int carryover= 0, produced, consumed, n;
  enum target_code target;

  if (argc != 2)
    help (*argv);

  if (!strcmp (argv[1], "gb") || !strcmp (argv[1], "GB"))
    target = gb;
  else if (!strcmp (argv[1], "cns") || !strcmp (argv[1], "CNS"))
    target = cns;
  else if (!strcmp (argv[1], "big5") || !strcmp (argv[1], "BIG5"))
    target = big5;
  else
    help (*argv);

  b5map_init ();
  initialize ();

  while ((n =
#ifdef INTERACTIVE
	  read (0, inbuf + carryover, INBUF_SIZE - carryover)
#else
	  fread (inbuf + carryover, 1, INBUF_SIZE - carryover, stdin)
#endif
	  )
	 > 0)
    {
      n += carryover;
      produced = decode (target, inbuf, n, outbuf, OUTBUF_SIZE, &consumed);
      fwrite (outbuf, 1, produced, stdout);
#ifdef INTERACTIVE
      fflush (stdout);
#endif
      carryover = n - consumed;
      for (n = 0; n < carryover; n++)
	inbuf[n] = inbuf[consumed + n];
    }

  if (carryover)
    {
      fwrite (inbuf, 1, carryover, stdout);
#ifdef INTERACTIVE
      fflush (stdout);
#endif
    }
  exit (0);
}
