/* en2022cn.c  -- ISO-2022-CN (RFC1922) encoder
   Copyright (C) 1995, 1996 Electrotechnical Laboratory, JAPAN.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


/* Author: K.Handa <handa@etl.go.jp>
   History:
	1996.4.6 -- Version 0.9beta released.
*/

#include <stdio.h>
#include "iso2022.h"

/* State of locking-shift.  0 means in SI, 1 means in SO.  */
int shiftout;

/* Table of <final-char> of character sets designated to each graphic
   registers.  The value -1 means nothing designated.  */
char REG[4];

/* Macros for indexing the array REG.  */
#define G0	0
#define G1	1
#define G2	2
#define G3	3

enum target_code {
  gb,
  cns,
  big5
};

/* Table of graphic registers which each character set (identified by
   <final-char>) be designated to.  */
char graphic_register[256];

/* Produce an ISO-2022 designation sequence for FINAL_CHAR in BUF.  */
#define DESIGNATION(buf, final_char) \
  *buf++ = ISO_ESC,\
  *buf++ = '$',\
  *buf++ = "()*+"[graphic_register[final_char]],\
  *buf++ = final_char,\
  REG[graphic_register[final_char]] = final_char

void
initialize ()
{
  int i;

  /* Initialize context of ISO-2022.  */
  shiftout = 0;
  REG[G0] = CHARSET_ASCII;
  REG[G1] = -1;
  REG[G2] = -1;
  REG[G3] = -1;

  /* Define where to designate each charcter sets. */
  for (i = 0; i < 256; i++)
    graphic_register[i] = -1;

  graphic_register[CHARSET_GB2312] = G1;
  graphic_register[CHARSET_GB7589] = G2;
  graphic_register[CHARSET_GB7590] = G3;

  graphic_register[CHARSET_GB12345] = G1;
  graphic_register[CHARSET_GB13131] = G2;
  graphic_register[CHARSET_GB13132] = G3;

  graphic_register[CHARSET_CNS11643_1] = G1;
  graphic_register[CHARSET_CNS11643_2] = G2;
  graphic_register[CHARSET_CNS11643_3] = G3;
  graphic_register[CHARSET_CNS11643_4] = G3;
  graphic_register[CHARSET_CNS11643_5] = G3;
  graphic_register[CHARSET_CNS11643_6] = G3;
  graphic_register[CHARSET_CNS11643_7] = G3;

  graphic_register[CHARSET_ISOIR165] = G3;
}

#define SAFETY_GET_SRC(c)	\
  do {			  	\
    if (src < src_end)	  	\
      c = *src++;		\
    else			\
      {			  	\
        src = src_base;	  	\
        goto no_more_src;	\
      }			  	\
  } while (0)

/* Encode SRC_LEN bytes GB-text (if TARGET == gb), CNS-text (if TARGET
   == cns) at SRC into ISO-2022-CN, or BIG5-text (if TARGET == big5).
   The resulting text goes to a place pointed by DST, the length of
   which should not exceed DST_LEN.  The bytes actually processed is
   returned as *CONSUMED.  The return value is the length of the
   encoded text.

   The format of GB-text is as follows:
   ----------------------------------------
   0xxxxxxx			 -- ASCII
   1xxxxxxx 1xxxxxxx		 -- GB2312
   0x8E 0xA2 1xxxxxxx 1xxxxxxx	 -- GB7589
   0x8E 0xA3 1xxxxxxx 1xxxxxxx	 -- GB7590

   The format of CNS-text is as follows:
   ----------------------------------------
   0xxxxxxx			 -- ASCII
   1xxxxxxx 1xxxxxxx		 -- CNS11643-1
   0x8E 0xA2 1xxxxxxx 1xxxxxxx	 -- CNS11643-2
   0x8E 0xA3 1xxxxxxx 1xxxxxxx	 -- CNS11643-3
   0x8E 0xA4 1xxxxxxx 1xxxxxxx	 -- CNS11643-4
   0x8E 0xA5 1xxxxxxx 1xxxxxxx	 -- CNS11643-5
   0x8E 0xA6 1xxxxxxx 1xxxxxxx	 -- CNS11643-6
   0x8E 0xA7 1xxxxxxx 1xxxxxxx	 -- CNS11643-7

   The format of BIG5-text is as follows:
   ----------------------------------------
   (1st byte)      0xA1 .. 0xFE
   (2nd byte)      0x40 .. 0x7E and 0xA1 .. 0xFE

 */

int
encode (target, src, src_len, dst, dst_len, consumed)
     enum target_code target;
     unsigned char *src, *dst;
     int src_len, dst_len;
     int *consumed;
{
  register unsigned int c1, c2, i;
  unsigned char *src_start = src, *src_end = src + src_len;
  unsigned char *dst_start = dst, *dst_end = dst + dst_len;

  /* We'll produce maximally 8 bytes in one loop.  So, subtract 7 from
     `dst_end' so that overflow check is necessary only at the head of
     loop.  */
  dst_end -= 7;

  while (src < src_end && dst < dst_end)
    {
      unsigned char *src_base = src;

      c1 = *src++;
      if (c1 < 0x80)		/* ASCII */
	{
	  if (shiftout)
	    {
	      *dst++ = ISO_SI;
	      shiftout = 0;
	    }
	  *dst++ = c1;
	  if (c1 == '\n')
	    /* Reset all designations at each end of line.  */
	    for (i = 1; i < 4; i++) REG[i] = -1;
	}
      else
	{
	  char charset;

	  if (target != big5)
	    {
	      if (c1 == ISO_SS2_8)
		{
		  SAFETY_GET_SRC (c1);
		  if (target == gb)
		    switch (c1)
		      {
		      case 0xA2: charset = CHARSET_GB7589; break;
		      case 0xA3: charset = CHARSET_GB7590; break;
		      default: goto invalid_code;
		      }
		  else		/* target == cns */
		    switch (c1)
		      {
		      case 0xA2: charset = CHARSET_CNS11643_2; break;
		      case 0xA3: charset = CHARSET_CNS11643_3; break;
		      case 0xA4: charset = CHARSET_CNS11643_4; break;
		      case 0xA5: charset = CHARSET_CNS11643_5; break;
		      case 0xA6: charset = CHARSET_CNS11643_6; break;
		      case 0xA7: charset = CHARSET_CNS11643_7; break;
		      default: goto invalid_code;
		      }
		  SAFETY_GET_SRC (c1);
		}
	      else
		charset = target == gb ? CHARSET_GB2312: CHARSET_CNS11643_1;

	      if (c1 <= 0xA0 || c1 == 0xFF) goto invalid_code;

	      SAFETY_GET_SRC (c2);
	      if (c2 <= 0xA0 || c2 == 0xFF) goto invalid_code;
	    }
	  else
	    {
	      unsigned int isocode;

	      if (c1 < 0xA1 || c1 > 0xFE) goto invalid_code;
	      SAFETY_GET_SRC (c2);
	      if (c2 < 0x40 || (c2 > 0x7E && c2 < 0xA1) || c2 > 0xFE)
		goto invalid_code;

	      BIG5_TO_ISO (c1, c2, charset, isocode);
	      if (!charset)
		{
		  charset = CHARSET_CNS11643_1;
		  isocode = BLANK_SQUARE_CNS11643_1;
		}
	      c1 = isocode >> 8, c2 = isocode & 0xFF;
	    }


	  if (REG[graphic_register[charset]] != charset)
	    DESIGNATION (dst, charset);

	  switch (graphic_register[charset])
	    {
	    case G1:		/* Use locking shift.  */
	      if (!shiftout)
		{
		  *dst++ = ISO_SO;
		  shiftout = 1;
		}
	      break;

	    case G2:		/* Use single shift 2.  */
	      *dst++ = ISO_ESC;
	      *dst++ = ISO_SS2_7;
	      break;

	    case G3:		/* Use single shift 3.  */
	      *dst++ = ISO_ESC;
	      *dst++ = ISO_SS3_7;
	      break;
	    }

	  *dst++ = c1 & 0x7F;
	  *dst++ = c2 & 0x7F;
	}
      continue;

    invalid_code:
      *dst++ = *src_base++;
      src = src_base;
    }

 no_more_src:
  *consumed = src - src_start;
  return (dst - dst_start);
}

void
help (cmd)
{
  fprintf (stderr, "Usage: %s TARGET_CODE < INFILE > OUTFILE\n", cmd);
  fprintf (stderr, "\tTARGET_CODE is one of `gb', `cns', or `big5'.\n");
  fprintf (stderr, "\tINFILE is a file of TARGET_CODE.\n");
  fprintf (stderr, "\tOUTFILE is a file encoded in ISO-2022-CN.\n");
  exit (1);
}

#define INBUF_SIZE 1024
#define OUTBUF_SIZE  1024

main (argc, argv)
     int argc;
     char *argv[];
{
  unsigned char inbuf[INBUF_SIZE];
  unsigned char outbuf[OUTBUF_SIZE];
  int carryover= 0, produced, consumed, n;
  enum target_code target;

  if (argc != 2)
    help (*argv);

  if (!strcmp (argv[1], "gb") || !strcmp (argv[1], "GB"))
    target = gb;
  else if (!strcmp (argv[1], "cns") || !strcmp (argv[1], "CNS"))
    target = cns;
  else if (!strcmp (argv[1], "big5") || !strcmp (argv[1], "BIG5"))
    {
      target = big5;
      b5map_init ();
    }
  else
    help (*argv);

  initialize ();

  while ((n = fread (inbuf + carryover, 1, INBUF_SIZE - carryover, stdin))
	 > 0)
    {
      n += carryover;
      produced = encode (target, inbuf, n, outbuf, OUTBUF_SIZE, &consumed);
      fwrite (outbuf, 1, produced, stdout);
      carryover = n - consumed;
      for (n = 0; n < carryover; n++)
	inbuf[n] = inbuf[consumed + n];
    }

  if (shiftout)
    {
      /* We must shift in at the end of file.  */
      outbuf[0] = ISO_SI;
      fwrite (outbuf, 1, 1, stdout);
    }
  if (carryover)
    fwrite (inbuf, 1, carryover, stdout);

  exit (0);
}
